import pytest
import json
import requests
from mock import patch, Mock
from wowpy.livestreams import LiveStream

@pytest.fixture
def live_stream_data():
    with open('tests/fixtures/spec_one_output_template_full.json') as json_file:
        specification = json.load(json_file)
    live_stream_data = {'live_stream': specification['live_stream']['parameters']}
    live_stream_data['live_stream']['name'] = specification['name']
    return live_stream_data

@patch('wowpy.livestreams.wowza_query')
def test_create_live_stream(mock_wowza_query, live_stream_data):
    mock_wowza_query.return_value = {'live_stream': {'id': 'abc123'}}
    live_stream_info = LiveStream.create_live_stream(data=live_stream_data)
    assert type(live_stream_info) is dict


from wowpy.livestreams import LiveStream
from wowpy.recordings import Recording
from wowpy.transcoders import Transcoder
from wowpy.constants import logger
from wowpy.exceptions import DeleteRecordingException

class Mapper:

  @classmethod
  def map_live_streams(cls):
    document_batch = []
    live_streams = LiveStream.get_live_streams()
    for live_stream in live_streams:
      live_stream_id = live_stream['id']
      live_stream_name = live_stream['name']
      document = {
        'live_stream_name': live_stream_name,
        'live_stream_id': live_stream_id
      }
      document_batch.append(document)
    return document_batch
    
  @classmethod
  def map_recordings(cls):
    """
    Live stream id has the same value that transcoder id
    """
    document_batch = []
    live_streams = LiveStream.get_live_streams()

    ### TODO: Find a better way
    live_stream_dict = {}
    for live_stream in live_streams:
      live_stream_dict[live_stream['id']] = live_stream['name']
    ###

    recordings = Recording.get_recordings()
    for recording in recordings:
      try:
        recording_id = recording['id']
        transcoder_id = recording['transcoder_id']
        live_stream_name = live_stream_dict[transcoder_id]
        document = {
          'transcoder_id': transcoder_id,
          'live_stream_name': live_stream_name,
          'recording_id': recording_id
        }
        document_batch.append(document)
      except KeyError as err:
        logger.warning('Not found %s', err)
    return document_batch



import dateutil.parser
from datetime import timedelta
from wowpy.constants import WSC_API_ENDPOINT, logger
from wowpy.utils import wowza_query, safe_run

class Schedule:

  scheduler_base = WSC_API_ENDPOINT + 'schedules'
  scheduler_single = scheduler_base + '/{scheduler_id}'  # get ,delete, update

  @classmethod
  def create_schedule(cls, data):
    response = wowza_query(endpoint=cls.scheduler_base, method='post', data=data)
    schedule = response['schedule']
    return schedule

  @classmethod
  def get_schedules(cls):
    # Get the list of schedules
    response = wowza_query(endpoint=cls.scheduler_base, method='get')
    logger.debug('Schedules response is {}'.format(response))
    schedules = response['schedules']
    return schedules

  @classmethod
  def get_schedule(cls, scheduler_id):
    endpoint = cls.scheduler_single.format(scheduler_id=scheduler_id)
    response = wowza_query(endpoint, 'get')
    schedule = response['schedule']
    return schedule

  @classmethod
  def delete_schedule(cls, scheduler_id):
    endpoint = cls.scheduler_single.format(scheduler_id=scheduler_id)
    response = wowza_query(endpoint, 'delete')
    schedule = response['schedule']
    return schedule

  @classmethod
  def update_schedule(cls, scheduler_id, data):
    endpoint = cls.scheduler_single.format(scheduler_id=scheduler_id)
    response = wowza_query(endpoint=endpoint, method='patch', data=data)
    schedule = response['schedule']
    return schedule

  @classmethod
  def process_schedule_data(cls, data, time_before=0, time_after=0):
    start_time = data['schedule']['start_transcoder']
    stop_time = data['schedule'].get('stop_transcoder', None)

    start_transcoder = dateutil.parser.parse(start_time) - timedelta(minutes=time_before)
    start_transcoder_iso = start_transcoder.strftime("%Y-%m-%dT%H:%M:%S.%fZ") # ISO 8601 datetime

    if stop_time:
      stop_transcoder = dateutil.parser.parse(stop_time) + timedelta(minutes=time_after)
    else: # not stop_transcoder specified
      stop_transcoder = dateutil.parser.parse(start_time) + timedelta(minutes=time_after)
    stop_transcoder_iso = stop_transcoder.strftime("%Y-%m-%dT%H:%M:%S.%fZ") # ISO 8601 datetime

    data['schedule']['start_transcoder'] = start_transcoder_iso
    data['schedule']['stop_transcoder'] = stop_transcoder_iso

    return data


from voluptuous import Schema, Required, All, Length, Range, Any, Optional

RESOURCE_SCHEMA = Schema({
    Required('version'): All(str, Length(min=0)),
    Required('name'): All(str, Length(min=0)),
    Required('id'): All(str, Length(min=0))
})

# TODO: Move all values to template, only left required ones

LIVE_STREAM_SCHEMA = Schema({
    Required('type'): 'live_stream',
    Required('id'): All(str, Length(min=0)),
    Required('parameters'): { # key is live_stream
        # Required
        Required('aspect_ratio_height'): All(int, Range(min=0)),
        Required('aspect_ratio_width'): All(int, Range(min=0)),
        Required('billing_mode'): All(str, Length(min=0)),
        Required('broadcast_location'): All(str, Length(min=0)),
        Required('encoder'): All(str, Length(min=0)),
        Required('name', default=''): All(str, Length(min=0)),
        Required('transcoder_type'): All(str, Length(min=0)),
        Required('delivery_method'): Any('pull','push'),
        # Non required
        'delivery_protocols': [All(str, Length(min=0)),All(str, Length(min=0))],
        'delivery_type': All(str, Length(min=0)),
        'disable_authentication': All(bool), # push
        'low_latency': All(bool),
        'recording': All(bool),
        'target_delivery_protocol': All(str, Length(min=0)),
        'use_stream_source': All(bool),
        'source_url': All(str, Length(min=0)),  # pull
    }
})

TRANSCODER_SCHEMA = Schema({
    Required('type'): 'transcoder', 
    Required('id'): All(str, Length(min=0)),
    Required('outputs'): All(
        [
            All(str, Length(min=0))
        ]
    ),
    Required('parameters'): { # key is empty
        Required('idle_timeout', default=0): All(int, Range(min=0))
    }
})

OUTPUT_SCHEMA = Schema({
    Required('type'): 'output',
    Required('id'): All(str, Length(min=0)),
    Required('targets'): All(
        [
            All(str, Length(min=0))
        ]
    ),
    Required('parameters'): {  # key is output
        # Required
        Required('stream_format'): All(str, Length(min=0)),
        Required('h264_profile'): All(str, Length(min=0)),
        Required('aspect_ratio_width'): All(int, Range(min=0)),
        Required('aspect_ratio_height'): All(int, Range(min=0)),
        Required('bitrate_video'): All(int, Range(min=0)),
        Required('bitrate_audio'): All(int, Range(min=0)),
        # Non required
        'framerate_reduction': All(str, Length(min=0)),
        'keyframes': All(str, Length(min=0)),
        'passthrough_audio': All(bool),
        'passthrough_video': All(bool)
    }
})

CUSTOM_TARGET_SCHEMA = Schema({
    Required('type'): 'custom_target',
    Required('id'): All(str, Length(min=0)),
    Required('parameters'): {  # key is stream_target_custom
        Required('provider'): Any('akamai_cupertino', 'rtmp', 'akamai_rtmp'),
        Required('primary_url'): All(str, Length(min=0)),
        Required('stream_name'): All(str, Length(min=0)),
        Required('use_https', default=True): All(bool)
    },
    'properties':
        All(
            [
                {
                    Required('property'): {
                        Required('key'): All(str, Length(min=0)),
                        Required('section'): All(str, Length(min=0)),
                        Required('value'): All(int, Range(min=0))
                    }
                }
            ]
        )
})



from wowpy.constants import WSC_API_ENDPOINT, logger
from wowpy.utils import wowza_query

class LiveStream:
  live_stream_base = WSC_API_ENDPOINT + 'live_streams'
  live_stream_pagination = live_stream_base + '/?page={page_number}&per_page=1000'
  live_stream_single = live_stream_base + '/{live_stream_id}'         # live_stream delete, get
  live_stream_single_state = live_stream_single + '/state'            # live_stream state
  live_stream_single_start = live_stream_single + '/start'            # live_stream start
  live_stream_single_stop = live_stream_single + '/stop'              # live_stream stop
  live_stream_single_stats = live_stream_single + '/stats'                  # live_stream stats
  live_stream_single_thumbnail_url = live_stream_single + '/thumbnail_url'  # live_stream thumbnail_url

  @classmethod
  def create_live_stream(cls, data):
    # Create wowza live stream 
    response = wowza_query(endpoint=cls.live_stream_base, method='post', data=data)
    return response['live_stream']

  @classmethod
  def update_live_stream(cls, live_stream_id, data):
    endpoint = cls.live_stream_single.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint=endpoint, method='patch', data=data)
    return response['live_stream']

  @classmethod
  def get_live_stream(cls, live_stream_id):
    # Get info of a live stream
    endpoint = cls.live_stream_single.format(
      live_stream_id=live_stream_id
    )
    response = wowza_query(endpoint=endpoint, method='get')
    live_stream = response['live_stream']
    logger.debug('Live stream info is {}'.format(live_stream))
    return live_stream

  @classmethod
  def delete_live_stream(cls, live_stream_id):
    # Delete livj1mjjpgne stream
    endpoint = cls.live_stream_single.format(
      live_stream_id=live_stream_id
    )
    wowza_query(endpoint=endpoint, method='delete')

  @classmethod
  def get_live_streams(cls):
    # Get the list of livestream
    response = wowza_query(endpoint=cls.live_stream_base, method='get')
    live_streams = response['live_streams']

    total_pages = response['pagination']['total_pages']
    for i in range(2, total_pages + 1):
      endpoint = cls.live_stream_pagination.format(page_number=i)
      response = wowza_query(endpoint=endpoint, method='get')
      live_streams.extend(response['live_streams'])
    return live_streams

  @classmethod
  def get_state(cls, live_stream_id):
    endpoint = cls.live_stream_single_state.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint, 'get')
    state = response['live_stream']['state']
    return state

  @classmethod
  def start_live_stream(cls, live_stream_id):
    endpoint = cls.live_stream_single_start.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint, 'put')
    state = response['live_stream']['state']
    return state
  
  @classmethod
  def stop_live_stream(cls, live_stream_id):
    endpoint = cls.live_stream_single_stop.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint, 'put')
    state = response['live_stream']['state']
    return state

  @classmethod
  def get_thumbnail_url(cls, live_stream_id):
    endpoint = cls.live_stream_single_thumbnail_url.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint, 'get')
    state = response['live_stream']['thumbnail_url']
    return state

  @classmethod
  def get_stats(cls, live_stream_id):
    endpoint = cls.live_stream_single_stats.format(live_stream_id=live_stream_id)
    response = wowza_query(endpoint, 'get')
    state = response['live_stream']
    return state


import pytest
import json
import requests
from unittest import skipIf
from mock import patch, Mock
from wowpy.resources import validate_resource, create_resource, update_resource
from wowpy.constants import SKIP_REAL

@pytest.fixture
def spec_one_output_template_full():
    with open('tests/fixtures/spec_one_output_template_full.json') as json_file:
        specification = json.load(json_file)
    return specification

@pytest.fixture
def spec_one_output_database():
    with open('tests/fixtures/spec_one_output_database.json') as json_file:
        specification = json.load(json_file)
    return specification

@pytest.fixture
def changes_spec_db_vs_spec_updated():
    with open('tests/fixtures/changes_spec_db_vs_spec_updated.json') as json_file:
        specification = json.load(json_file)
    return specification

def test_validate_resource(spec_one_output_template_full):
    valid = validate_resource(specification=spec_one_output_template_full)
    assert valid == True

@patch('wowpy.resources.Transcoder.associate_target_stream')
@patch('wowpy.resources.TargetStream.update_properties')
@patch('wowpy.resources.TargetStream.create_target')
@patch('wowpy.resources.Transcoder.create_transcoder_output') # returns an output id
@patch('wowpy.resources.Transcoder.update_transcoder')
@patch('wowpy.resources.Transcoder.get_transcoder_outputs')
@patch('wowpy.resources.LiveStream.create_live_stream')
def test_create_resource(mock_create_live,
                         mock_get_trans,
                         mock_update_trans, 
                         mock_create_output, 
                         mock_create_target, 
                         mock_update_target, 
                         mock_target_associate, 
                         spec_one_output_template_full):
    mock_create_live.return_value = {'id': 'abc123'}
    mock_get_trans.return_value = []
    mock_update_trans.return_value = None
    mock_create_output.return_value = 'abc123'
    mock_create_target.return_value = 'abc-123'
    mock_update_target.return_value = None
    mock_update_trans.return_value = None
    mock_target_associate.return_value = None
    specification = create_resource(specification=spec_one_output_template_full)
    assert type(specification) is dict

@patch('wowpy.resources.Transcoder.associate_target_stream')
@patch('wowpy.resources.TargetStream.update_properties')
@patch('wowpy.resources.TargetStream.create_target')
def test_update_resource(mock_create_target, 
                         mock_update_target, 
                         mock_target_associate,
                         spec_one_output_database,
                         changes_spec_db_vs_spec_updated):
    mock_create_target.return_value = 'abc-123'
    mock_update_target.return_value = None
    mock_target_associate.return_value = None
    updated_spec_with_ids = update_resource(spec_one_output_database, changes_spec_db_vs_spec_updated)
    assert type(updated_spec_with_ids) is dict

###

@skipIf(SKIP_REAL, 'Skipping tests that hit the real API server.')
def test_real_create_resource(spec_one_output_template_full):
    specification = create_resource(specification=spec_one_output_template_full)
    assert type(specification) is dict


from wowpy.constants import WSC_API_ENDPOINT, logger
from wowpy.utils import wowza_query
from wowpy.constants import logger
from wowpy.exceptions import DeleteRecordingException

class Recording:
  recording_base = WSC_API_ENDPOINT + 'recordings'
  recording_pagination = recording_base + '/?page={page_number}&per_page=1000'
  recording_single = recording_base + '/{recording_id}'                         # recording delete, get

  @classmethod
  def get_recording(cls, recording_id):
    # Get recording info
    endpoint = cls.recording_single.format(
      recording_id=recording_id
    )
    response = wowza_query(endpoint=endpoint, method='get')
    recording = response['recording']
    logger.debug('Recording info is {}'.format(recording))
    return recording

  @classmethod
  def get_recordings(cls):
    # Get the list of livestream
    response = wowza_query(endpoint=cls.recording_base, method='get')
    recordings = response['recordings']

    total_pages = response['pagination']['total_pages']
    for i in range(2, total_pages + 1):
      try:
        endpoint = cls.recording_pagination.format(page_number=i)
        response = wowza_query(endpoint=endpoint, method='get')
        recordings.extend(response['recordings'])
      except DeleteRecordingException as err:
        logger.warning(err)
    return recordings


import setuptools

with open('VERSION', 'r') as fh:
    version = fh.read().strip()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wowpy",
    version=version,
    install_requires=[
      "setuptools",
      "requests",
      "voluptuous"
    ],
    author="Daniel Barragan",
    author_email="dbarragan1331@gmail.com",
    description="Wowza python wrapper - wowpy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/d4n13lbc/wowpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={'': ['config.json']},
    include_package_data=True,
    zip_safe=False
)


from wowpy.livestreams import LiveStream
from wowpy.transcoders import Transcoder
from wowpy.targets import TargetStream
from wowpy.utils import validate_schema, get_operation
from wowpy.constants import logger
from wowpy import schemas

def create_resource(specification):

  # Get data from specification
  version = specification['version']
  resource_name = specification['name']
  resource_keys = list(specification.keys())

  for resource_key in resource_keys:

    item = specification[resource_key]

    if type(item) is not dict:
      continue

    resource_type = item.get('type', None)
    if resource_type in [None, 'custom_target', 'output']:
      continue

    if resource_type == 'live_stream':
      live_stream_data = {'live_stream': specification[resource_key]['parameters']}
      live_stream_data['live_stream']['name'] = resource_name
      live_stream = LiveStream.create_live_stream(data=live_stream_data)
      live_stream_id = live_stream['id']
      outputs = Transcoder.get_transcoder_outputs(transcoder_id=live_stream_id)
      specification[resource_key]['id'] = live_stream_id
      specification['id'] = live_stream_id # set global id

      # Delete default output
      for output in outputs:
        output_id = output['id']
        Transcoder.delete_transcoder_output(transcoder_id=live_stream_id, output_id=output_id)
        target_id = output['output_stream_targets'][0]['stream_target']['id']
        TargetStream.delete_target(stream_type='fastly', stream_target_id=target_id)

    if resource_type == 'transcoder':
      transcoder_data = {'transcoder': specification[resource_key]['parameters']} 
      Transcoder.update_transcoder(transcoder_id=live_stream_id, data=transcoder_data)
      specification[resource_key]['id'] = live_stream_id

      output_keys = specification[resource_key]['outputs']
      for output_key in output_keys:
        output_data = {'output': specification[output_key]['parameters']}
        output_id = Transcoder.create_transcoder_output(transcoder_id=live_stream_id, data=output_data)
        specification[output_key]['id'] = output_id

        target_keys = specification[output_key]['targets']
        for target_key in target_keys:
          target_data = {'stream_target_custom': specification[target_key]['parameters']}
          target_provider = specification[target_key]['parameters']['provider'].split('_')
          target_provider_upper = target_provider[0].upper()
          target_name = '-'.join([resource_name, target_provider_upper])
          target_data['stream_target_custom']['name'] = target_name
          target_id = TargetStream.create_target(data=target_data)
          target_properties = specification[target_key]['properties']

          TargetStream.update_properties(stream_target_id=target_id, properties_data=target_properties)
          Transcoder.associate_target_stream(transcoder_id=live_stream_id, output_id=output_id, stream_target_id=target_id)
          specification[target_key]['id'] = target_id

  return specification
    
def update_resource(stored_spec, changes):
  resource_name = stored_spec['name']
  live_stream_id = stored_spec['id']
  changed_keys = list(changes.keys())

  targets = []
  outputs = []

  for changed_key in changed_keys:
    stored_data = stored_spec.get(changed_key, None) 
    if stored_data:
      resource_type = stored_data['type']
      resource_id = stored_data['id']
      resource_key = changed_key
    else:
      resource_type = changes[changed_key]['type']
      resource_id = ''
      resource_key = changed_key

    if resource_type == 'custom_target':
      resource_data = changes[changed_key]
      parameters = resource_data.get('parameters', None)
      properties = resource_data.get('properties', [])
      targets.append(
        {
          'id': resource_id,
          'type': resource_type,
          'key': resource_key,
          'parameters': parameters,
          'properties': properties
        }
      )

    if resource_type == 'output':
      resource_data = changes[changed_key]
      parameters = resource_data.get('parameters', None)
      output_targets = resource_data.get('targets', [])
      outputs.append(
        {
          'id': resource_id,
          'type': resource_type,
          'key': resource_key,
          'parameters': parameters,
          'targets': output_targets
        }
      )

  for target in targets:
    operation = get_operation(data=target)

    if operation == 'new':
      target_key = target['key']
      target_data = {'stream_target_custom': target['parameters']} 
      target_provider = target['parameters']['provider'].split('_')
      target_provider_upper = target_provider[0].upper()
      target_name = '-'.join([resource_name, target_provider_upper])
      # With the following line the 'name' key is added to target['parameters'] 
      target_data['stream_target_custom']['name'] = target_name
      target_id = TargetStream.create_target(data=target_data)
      target_properties = target['properties']
      TargetStream.update_properties(stream_target_id=target_id, properties_data=target_properties)
      stored_spec[target_key] = {
        'id': target_id,
        'type': target['type'],
        'parameters': target['parameters'],
        'properties': target['properties']
      }

  for output in outputs:
    operation = get_operation(data=output)
    output_key = output['key']
    if operation == 'insert':
      new_targets = output['targets']['$insert']
      for new_target in new_targets:
        target_key = new_target[1]
        output_id = output['id']
        target_id = stored_spec[target_key]['id']
        Transcoder.associate_target_stream(transcoder_id=live_stream_id, output_id=output_id, stream_target_id=target_id)
        stored_spec[output_key]['targets'].append(target_key)

  return stored_spec

def delete_resource(live_stream_id):
  try: 
    live_stream_info = LiveStream.get_live_stream(live_stream_id=live_stream_id)
    LiveStream.delete_live_stream(live_stream_id=live_stream_id)
    stream_targets = live_stream_info['stream_targets']
    for stream_target in stream_targets:
      stream_target_id = stream_target['id']
      TargetStream.delete_target('custom', stream_target_id)
  except Exception as excp:
    logger.info('Resource {live_stream_id} not propertly deleted'.format(live_stream_id=live_stream_id))
    raise Exception(str(excp))

# TODO: Validate lists in resource blocks

def validate_resource(specification):
  block_keys = specification.keys()
  for block_key in block_keys:
    if block_key in ['version', 'name', 'id']:
      continue
    data_block = specification[block_key]
    schema_name = data_block['type']+'_'+'schema'
    schema_uppercase = schema_name.upper().replace(' ','_') # upper replaces underscore with space
    schema_model = getattr(schemas, schema_uppercase)
    response = validate_schema(schema=schema_model, data=data_block)
    if not response['valid']:
      return False
  return True

def get_resource_info(id):
  resource_info = {
    'livestream': {},
    'transcoder': {},
    'targets': [],
    'schedule': {}
  }

  live_stream_info = LiveStream.get_live_stream(id)
  resource_info['livestream'] = live_stream_info
  transcoder_info = Transcoder.get_transcoder(id)
  resource_info['transcoder'] = transcoder_info
  transcoder_outputs = transcoder_info['outputs']
  for output_item in transcoder_outputs:
    stream_targets = output_item['output_stream_targets']
    for target_item in stream_targets:
      stream_target_id = target_item['stream_target']['id']
      stream_target_type = target_item['stream_target']['type']
      stream_target_info = TargetStream.get_target(stream_type=stream_target_type, stream_target_id=stream_target_id)
      stream_target_properties = TargetStream.get_target_properties(stream_type=stream_target_type, stream_target_id=stream_target_id)
      resource_info['targets'].append({'config': stream_target_info, 'properties': stream_target_properties})

#   schedule_info = Schedule.get_schedule(scheduler_id=scheduler_id)
#   resource_info['schedule'] = schedule_info

  return resource_info